module doodul.lobby;

import std.algorithm : find, filter, map, remove, any;
import std.array : array;
import std.conv : to;

import vibe.vibe;

import doodul.app;
import doodul.game;
import doodul.player;

Lobby[] lobbies;

void lobbyGet(HTTPServerRequest request, HTTPServerResponse response)
{
	auto lobbyId = request.params["id"];

	response.redirect("/?lobbyId=" ~ lobbyId.urlEncode);
}

void lobbyPost(HTTPServerRequest request, HTTPServerResponse response)
{
	auto lobbyId = request.params["id"];
	auto nickname = request.form["nickname"];

	response.render!("lobby.dt", lobbyId, nickname);
}

void getLobbyJoinability(HTTPServerRequest request, HTTPServerResponse response)
{
	struct JoinabilityResponse
	{
		bool canJoin;
		string badField;
		string reason;
	}

	auto nickname = request.query["nickname"];
	if(nickname.empty)
	{
		response.writeJsonBody(JoinabilityResponse(false, "nickname", "Please enter a nickname."));
		return;
	}
	auto lobbyId = request.params["id"];
	if(lobbyId.empty)
	{
		response.writeJsonBody(JoinabilityResponse(false, "lobbyId", "Please enter a lobby id."));
		return;
	}

	auto lobbyRange = lobbies.find!(l => l.id == lobbyId);
	if(!lobbyRange.empty)
	{
		auto lobby = lobbyRange[0];
		if(lobby.players.length >= 8) //TODO: Make this configurable
		{
			response.writeJsonBody(JoinabilityResponse(false, "lobbyId", "This lobby is currently full."));
			return;
		}
		else if(lobby.players.any!(p => p.nickname == nickname)) //TODO: Maybe do case insensitive compare?
		{
			response.writeJsonBody(JoinabilityResponse(false, "nickname", "There is someone with this nickname in the lobby already."));
			return;
		}
	}

	response.writeJsonBody(JoinabilityResponse(true));
}

void wsLobby(scope WebSocket socket)
{
	Player player;
	Lobby lobby;

	auto sendLobbyInfo = delegate(Player p) {
		struct LobbyInfo
		{
			string leader;
			string[] players;
			int numberOfRounds;
		}
		auto lobbyInfo = LobbyInfo(lobby.leader.nickname, lobby.players.map!(p => p.nickname).array, lobby.numberOfRounds);
		p.sendMessage("setLobbyInfo", lobbyInfo.serializeToJson);
	};

	while(socket.connected)
	{
		string messageText;
		try
		{
			messageText = socket.receiveText();
		}
		catch(Exception exception)
		{
			break;
		}
		auto message = deserializeJson!WebSocketMessage(messageText);

		switch(message.type)
		{
			case "joinLobby":
				//TODO: Fix race condition (nickname collision)
				auto nickname = message.data.get!string;

				player = new Player(nickname, socket);

				auto lobbyId = socket.request.params["id"];
				auto lobbyRange = lobbies.find!(l => l.id == lobbyId);
				if(lobbyRange.empty)
				{
					lobby = new Lobby(lobbyId, player);
					lobbies ~= lobby;
				}
				else
				{
					lobby = lobbyRange[0];
				}
				lobby.addPlayer(player);

				if(lobby.game !is null)
				{
					player.sendMessage("startGame", null.serializeToJson);
					player.socket = null;
					continue;
				}

				foreach(p; lobby.players)
				{
					sendLobbyInfo(p);
				}
				break;
			//TODO: Checjk if game is started
			case "setNumberOfRounds":
				if(lobby is null || player != lobby.leader)
				{
					//TODO: Send error message?
					continue;
				}

				lobby.numberOfRounds = message.data.get!string.to!int;

				foreach(p; lobby.players.filter!(p => p != lobby.leader))
				{
					sendLobbyInfo(p);
				}
				break;
			case "startGame":
				if(lobby is null || player != lobby.leader || lobby.players.length < 2)
				{
					//TODO: Send error message?
					continue;
				}

				lobby.game = new Game(lobby, lobby.numberOfRounds);

				foreach(p; lobby.players)
				{
					p.sendMessage("startGame", null.serializeToJson);
					p.socket = null;
				}
				break;
			default:
				logWarn("Unknown WebSocket message type: " ~ message.type);
				break;
		}
	}

	if(lobby is null || lobby.game !is null)
	{
		return;
	}

	lobby.players = lobby.players.remove!(p => p == player);
	if(lobby.players.length == 0)
	{
		lobbies = lobbies.remove!(l => l == lobby);
		return;
	}

	if(lobby.leader == player)
	{
		lobby.leader = lobby.players[0];
	}

	foreach(p; lobby.players)
	{
		sendLobbyInfo(p);
	}
}

class Lobby
{
	string id;

	Player leader;
	Player[] players;

	int numberOfRounds = 3;
	//TODO: Word list member

	Game game;

	this(string id, Player leader)
	{
		this.id = id;
		this.leader = leader;
	}

	void addPlayer(Player player)
	{
		players ~= player;
	}
}
